﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Core;
using HeuristicLab.Parameters;
using HeuristicLab.Common;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Data;

namespace HeuristicLab.Problems.MetaOptimization {
  // todo: item name/descr...
  [StorableClass]
  public class ParameterConfigurationManipulator : SingleSuccessorOperator, IParameterConfigurationManipulator, IStochasticOperator {
    public override bool CanChangeName {
      get { return false; }
    }
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }
    public IValueLookupParameter<ParameterConfigurationTree> ParameterConfigurationTreeParameter {
      get { return (IValueLookupParameter<ParameterConfigurationTree>)Parameters["ParameterConfigurationTree"]; }
    }

    public IValueLookupParameter<IIntValueManipulator> IntValueManipulatorParameter {
      get { return (IValueLookupParameter<IIntValueManipulator>)Parameters[MetaOptimizationProblem.IntValueManipulatorParameterName]; }
    }
    public IValueLookupParameter<IDoubleValueManipulator> DoubleValueManipulatorParameter {
      get { return (IValueLookupParameter<IDoubleValueManipulator>)Parameters[MetaOptimizationProblem.DoubleValueManipulatorParameterName]; }
    }

    public ParameterConfigurationManipulator() {
      Parameters.Add(new ValueLookupParameter<IRandom>("Random", "The pseudo random number generator which should be used for stochastic manipulation operators."));
      Parameters.Add(new ValueLookupParameter<ParameterConfigurationTree>("ParameterConfigurationTree", "The parameter configuration which should be manipulated."));

      Parameters.Add(new ValueLookupParameter<IIntValueManipulator>(MetaOptimizationProblem.IntValueManipulatorParameterName, ""));
      Parameters.Add(new ValueLookupParameter<IDoubleValueManipulator>(MetaOptimizationProblem.DoubleValueManipulatorParameterName, ""));
    }
    [StorableConstructor]
    protected ParameterConfigurationManipulator(bool deserializing) : base(deserializing) { }
    protected ParameterConfigurationManipulator(ParameterConfigurationManipulator original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new ParameterConfigurationManipulator(this, cloner);
    }

    public sealed override IOperation Apply() {
      Apply(RandomParameter.ActualValue, ParameterConfigurationTreeParameter.ActualValue, IntValueManipulatorParameter.ActualValue, DoubleValueManipulatorParameter.ActualValue);
      return base.Apply();
    }

    public static void Apply(IRandom random, IValueConfiguration configuration, IIntValueManipulator intValueManipulator, IDoubleValueManipulator doubleValueManipulator) {
      configuration.Mutate(random, Mutate, intValueManipulator, doubleValueManipulator);
    }

    private static void Mutate(IRandom random, IOptimizable configuration, IIntValueManipulator intValueManipulator, IDoubleValueManipulator doubleValueManipulator) {
      var vc = configuration as IValueConfiguration;
      var pc = configuration as IParameterConfiguration;

      if (vc != null) {
        var value = vc.ActualValue.Value;
        var range = vc.RangeConstraint;
        if (value is IntValue) {
          intValueManipulator.Apply(random, (IntValue)value, (IntValueRange)range);
        } else if (value is PercentValue) {
          doubleValueManipulator.Apply(random, (PercentValue)value, ((PercentValueRange)range).AsDoubleValueRange());
        } else if (value is DoubleValue) {
          doubleValueManipulator.Apply(random, (DoubleValue)value, (DoubleValueRange)range);
        }
      } else if (pc != null) {
        do {
          pc.ActualValueConfigurationIndex = random.Next(pc.ValueConfigurations.Count());
        } while (!pc.ValueConfigurations.ItemChecked(pc.ValueConfigurations[pc.ActualValueConfigurationIndex]));
        pc.ActualValue = pc.ValueConfigurations[pc.ActualValueConfigurationIndex].ActualValue;
      }
    }
  }
}